from django.core.exceptions import ObjectDoesNotExist
from django.db import connections
from django.db.utils import IntegrityError

from arkindex.documents.models import Entity, EntityType, MLClass, TranscriptionEntity
from arkindex.documents.tasks import worker_results_delete
from arkindex.process.models import ProcessMode, WorkerVersion
from arkindex.project.tests import FixtureTestCase
from arkindex.training.models import DatasetSet, Model, ModelVersionState


class TestDeleteWorkerResults(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.text = MLClass.objects.create(name="text", corpus=cls.corpus)
        cls.version_1 = WorkerVersion.objects.get(worker__slug="reco")
        cls.version_2 = WorkerVersion.objects.get(worker__slug="dla")
        cls.process = cls.version_1.processes.get(mode=ProcessMode.Workers)
        cls.model = Model.objects.create(name="Generic model", public=False)
        cls.model_version = cls.model.versions.create(
            state=ModelVersionState.Available,
            tag="Test",
            hash="A" * 32,
            archive_hash="42",
            size=1337,
        )
        cls.configuration = cls.version_2.worker.configurations.create(
            name="DLA configuration", configuration={"value": "test"}
        )
        cls.process_2 = cls.user.processes.create(
            mode=ProcessMode.Workers,
            corpus=cls.corpus,
            name="Second process",
        )
        cls.worker_run_1 = cls.version_1.worker_runs.get(process=cls.process)
        cls.worker_run_2 = cls.version_1.worker_runs.create(
            process=cls.process_2,
            version=cls.version_1,
            model_version=cls.model_version,
            configuration=cls.configuration,
            parents=[],
        )

        cls.vol = cls.corpus.elements.get(name="Volume 1")
        cls.classification1 = cls.vol.classifications.create(
            worker_version=cls.version_1,
            worker_run=cls.worker_run_1,
            ml_class=cls.text,
            confidence=.42
        )

        cls.page1 = cls.corpus.elements.get(name="Volume 1, page 1r")
        cls.classification2 = cls.page1.classifications.create(
            worker_version=cls.version_1,
            worker_run=cls.worker_run_1,
            ml_class=cls.text,
            confidence=.42
        )
        cls.transcription1 = cls.page1.transcriptions.create(
            text="something",
            worker_version=cls.version_1,
            worker_run=cls.worker_run_2,
            confidence=0.42,
        )
        cls.person_type = EntityType.objects.get(
            name="person",
            corpus=cls.corpus
        )
        cls.entity = Entity.objects.create(
            type=cls.person_type,
            corpus=cls.corpus,
            name="entity 1",
            worker_version=cls.version_2,
            worker_run=cls.worker_run_2,
        )
        cls.transcription_entity1 = TranscriptionEntity.objects.create(
            transcription=cls.transcription1,
            entity=cls.entity,
            offset=2,
            length=len(cls.entity.name),
            worker_version=cls.version_1,
            worker_run=cls.worker_run_2,
        )

        cls.page2 = cls.corpus.elements.get(name="Volume 1, page 2r")
        cls.page2.worker_version = cls.version_1
        cls.page2.save()
        cls.classification3 = cls.page2.classifications.create(
            ml_class=cls.text,
            confidence=.42,
            worker_version=cls.version_2,
            worker_run=cls.worker_run_1,
        )
        cls.transcription2 = cls.page2.transcriptions.create(
            text="something",
            confidence=0.42,
            worker_version=cls.version_1,
            worker_run=cls.worker_run_2,
        )
        cls.transcription_entity2 = TranscriptionEntity.objects.create(
            transcription=cls.transcription2,
            entity=cls.entity,
            offset=2,
            length=len(cls.entity.name),
            worker_version=cls.version_1,
            worker_run=cls.worker_run_2,
        )

    def check_deleted(self, *deleted_items):
        """
        Helper function to check which items are deleted and which items are preserved
        """
        items = (self.vol, self.page1, self.page2, self.classification1, self.classification2, self.classification3, self.transcription1, self.transcription2, self.entity, self.transcription_entity1, self.transcription_entity2)
        expected_del_ids = tuple(i.id for i in deleted_items)

        non_deleted = []
        missing = []
        for item in items:
            try:
                item.refresh_from_db()
            except ObjectDoesNotExist:
                if item.id not in expected_del_ids:
                    missing.append(item)
            else:
                if item.id in expected_del_ids:
                    non_deleted.append(item)
        self.assertFalse(
            bool(non_deleted or missing),
            f"Missing items: {missing}. Non deleted items: {non_deleted}.",
        )

    def test_run_on_corpus(self):
        with self.assertExactQueries("worker_results_delete_in_corpus.sql", params={
            "corpus_id": str(self.corpus.id),
            "version_id": str(self.version_1.id),
        }):
            worker_results_delete(
                corpus_id=self.corpus.id,
                version_id=self.version_1.id,
            )
        self.check_deleted(
            self.classification1,
            self.classification2,
            self.classification3,
            self.transcription1,
            self.transcription2,
            self.transcription_entity1,
            self.transcription_entity2,
            self.page2,
        )

    def test_run_on_parent(self):
        with self.assertExactQueries("worker_results_delete_under_parent.sql", params={
            "corpus_id": str(self.corpus.id),
            "version_id": str(self.version_1.id),
            "element_id": str(self.page1.id),
        }):
            worker_results_delete(corpus_id=self.corpus.id, version_id=self.version_1.id, element_id=self.page1.id)
        self.check_deleted(
            self.classification2,
            self.transcription1,
            self.transcription_entity1,
        )

    def test_run_on_parent_delete_element(self):
        """
        The element itself is deleted after its related results from the same version
        """
        with self.assertExactQueries("worker_results_delete_under_parent_included.sql", params={
            "corpus_id": str(self.corpus.id),
            "version_id": str(self.version_1.id),
            "element_id": str(self.page2.id),
        }):
            worker_results_delete(corpus_id=self.corpus.id, version_id=self.version_1.id, element_id=self.page2.id)
        self.check_deleted(
            self.classification3,
            self.transcription2,
            self.transcription_entity2,
            self.page2,
        )

    def test_run_model_version_filter_on_parent(self):
        with self.assertExactQueries("worker_results_delete_model_version_under_parent.sql", params={
            "corpus_id": str(self.corpus.id),
            "element_id": str(self.page2.id),
            "model_version_id": str(self.model_version.id),
        }):
            worker_results_delete(
                corpus_id=self.corpus.id,
                element_id=self.page2.id,
                model_version_id=self.model_version.id,
            )
        self.check_deleted(
            self.transcription2,
            self.transcription_entity2,
        )

    def test_run_configuration_filter(self):
        with self.assertExactQueries("worker_results_delete_configuration_filter.sql", params={
            "corpus_id": str(self.corpus.id),
            "configuration_id": str(self.configuration.id),
        }):
            worker_results_delete(
                corpus_id=self.corpus.id,
                configuration_id=self.configuration.id,
            )
        self.check_deleted(
            self.transcription1,
            self.transcription2,
            self.transcription_entity1,
            self.transcription_entity2,
            self.entity,
        )

    def test_run_no_configuration_filter(self):
        # Results generated with worker run 2 are preserved
        self.page2.worker_run = self.worker_run_2
        self.page2.save()
        with self.assertExactQueries("worker_results_delete_unset_configuration.sql", params={
            "corpus_id": str(self.corpus.id),
            "element_id": str(self.page2.id),
        }):
            worker_results_delete(
                corpus_id=self.corpus.id,
                element_id=self.page2.id,
                configuration_id=False,
            )
        self.check_deleted(
            self.classification3,
        )

    def test_run_model_version_configuration_worker_version_filter(self):
        """
        Model version, configuration and worker version can be used simultaneously
        """
        worker_results_delete(
            corpus_id=self.corpus.id,
            version_id=self.version_1.id,
            model_version_id=self.model_version.id,
            configuration_id=self.configuration.id,
        )
        self.check_deleted(
            self.transcription1,
            self.transcription_entity1,
            self.transcription2,
            self.transcription_entity2,
        )

    def test_run_all_versions(self):
        with self.assertExactQueries("worker_results_delete_all_versions.sql", params={
            "corpus_id": str(self.corpus.id),
        }):
            worker_results_delete(corpus_id=self.corpus.id)
        self.check_deleted(
            self.classification1,
            self.classification2,
            self.classification3,
            self.transcription1,
            self.transcription2,
            self.transcription_entity1,
            self.transcription_entity2,
            self.page2,
            self.entity,
        )

    def test_run_dataset_failure(self):
        """
        Elements that are part of a dataset cannot be deleted
        """
        self.page1.worker_run = self.worker_run_1
        self.page1.worker_version = self.version_1
        self.page1.save()
        DatasetSet.objects.get(name="test", dataset__name="First Dataset").set_elements.create(element=self.page1)

        worker_results_delete(corpus_id=self.corpus.id)
        # Prevent delaying constraints check at end of the test transaction
        # by directly setting "SET CONSTRAINTS ALL IMMEDIATE"
        # https://code.djangoproject.com/ticket/11665
        with self.assertRaises(IntegrityError):
            connections["default"].check_constraints()

    def test_run_worker_run_or_version(self):
        with self.assertRaisesMessage(AssertionError, "The worker_run_id and version_id parameters are mutually exclusive."):
            worker_results_delete(
                corpus_id=self.corpus.id,
                version_id=self.version_1.id,
                worker_run_id=self.worker_run_1.id
            )

    def test_run_worker_run_on_corpus(self):
        with self.assertExactQueries("worker_results_delete_in_corpus_worker_run.sql", params={
            "corpus_id": str(self.corpus.id),
            "worker_run_id": str(self.worker_run_1.id),
        }):
            worker_results_delete(
                corpus_id=self.corpus.id,
                worker_run_id=self.worker_run_1.id,
            )
        self.check_deleted(
            self.classification1,
            self.classification2,
            self.classification3
        )

    def test_run_worker_run_on_parent(self):
        with self.assertExactQueries("worker_results_delete_under_parent_worker_run.sql", params={
            "corpus_id": str(self.corpus.id),
            "worker_run_id": str(self.worker_run_2.id),
            "element_id": str(self.page1.id),
        }):
            worker_results_delete(corpus_id=self.corpus.id, worker_run_id=self.worker_run_2.id, element_id=self.page1.id)
        self.check_deleted(
            self.transcription1,
            self.transcription_entity1,
        )

    def test_run_worker_run_on_parent_delete_element(self):
        """
        The element itself is deleted after its related results from the same worker run
        """
        self.page1.worker_run = self.worker_run_2
        self.page1.worker_version = self.version_2
        self.page1.save()
        with self.assertExactQueries("worker_results_delete_under_parent_included_worker_run.sql", params={
            "corpus_id": str(self.corpus.id),
            "worker_run_id": str(self.worker_run_2.id),
            "element_id": str(self.page1.id),
        }):
            worker_results_delete(corpus_id=self.corpus.id, worker_run_id=self.worker_run_2.id, element_id=self.page1.id)
        self.check_deleted(
            self.transcription1,
            self.transcription_entity1,
            self.page1,
            # self.classifications2 is deleted as well since it's on self.page1
            self.classification2
        )

    def test_run_worker_run_ignore_filters(self):
        with self.assertExactQueries("worker_results_delete_in_corpus_worker_run.sql", params={
            "corpus_id": str(self.corpus.id),
            "worker_run_id": str(self.worker_run_1.id)
        }):
            worker_results_delete(
                corpus_id=self.corpus.id,
                worker_run_id=self.worker_run_1.id,
                model_version_id=self.model_version.id,
                configuration_id=self.configuration.id
            )
        self.check_deleted(
            self.classification1,
            self.classification2,
            self.classification3
        )
